import { ServerResponse } from 'node:http';
import { BucketItemStat } from 'minio';
import fs from 'node:fs';
import path from 'node:path';

const viewableExtensions = ['jpg', 'jpeg', 'png', 'gif', 'svg', 'webp', 'mp4', 'webm', 'mp3', 'wav', 'ogg', 'pdf', 'doc', 'docx', 'xls', 'xlsx', 'ppt', 'pptx'];
import { OssBase } from '../index.ts';
/**
 * 过滤 metaData 中的 key, 去除 password, accesskey, secretkey，
 * 并返回过滤后的 metaData
 * @param metaData
 * @returns
 */
export const filterMetaDataKeys = (metaData: Record<string, string>, clearKeys: string[] = []) => {
  const keys = Object.keys(metaData);
  // remove X-Amz- meta data
  const removeKeys = ['password', 'accesskey', 'secretkey', ...clearKeys];
  const filteredKeys = keys.filter((key) => !removeKeys.includes(key));
  return filteredKeys.reduce((acc, key) => {
    acc[key] = metaData[key];
    return acc;
  }, {} as Record<string, string>);
};
type SendObjectOptions = {
  res: ServerResponse;
  client: OssBase;
  objectName: string;
  isDownload?: boolean;
};
export const NotFoundFile = (res: ServerResponse, msg?: string, code = 404) => {
  res.writeHead(code, { 'Content-Type': 'text/plain' });
  res.end(msg || 'Not Found File');
  return;
};
export const sendObject = async ({ res, objectName, client, isDownload = false }: SendObjectOptions) => {
  let stat: BucketItemStat;
  try {
    stat = await client.statObject(objectName);
  } catch (e) {
  } finally {
    if (!stat || stat.size === 0) {
      return NotFoundFile(res);
    }
    const contentLength = stat.size;
    const etag = stat.etag;
    const lastModified = stat.lastModified.toISOString();
    const filename = objectName.split('/').pop() || 'no-file-name-download'; // Extract filename from objectName
    const fileExtension = filename.split('.').pop()?.toLowerCase() || '';
    const filteredMetaData = filterMetaDataKeys(stat.metaData, ['size', 'etag', 'last-modified']);
    const contentDisposition = viewableExtensions.includes(fileExtension) && !isDownload ? 'inline' : `attachment; filename="${filename}"`;

    res.writeHead(200, {
      'Content-Length': contentLength,
      etag,
      'last-modified': lastModified,
      'Content-Disposition': contentDisposition,
      'file-name': filename,
      ...filteredMetaData,
    });
    const objectStream = await client.getObject(objectName);

    objectStream.pipe(res, { end: true });
  }
};

export const downloadObject = async ({ objectName, client, filePath }: Pick<SendObjectOptions, 'objectName' | 'client'> & { filePath: string }) => {
  const objectStream = await client.getObject(objectName);
  const dir = path.dirname(filePath);
  if (!fs.existsSync(dir)) {
    fs.mkdirSync(dir, { recursive: true });
  }
  objectStream.pipe(fs.createWriteStream(filePath));
  return objectStream;
};
